import numpy as np
import sys
# import warnings
# warnings.filterwarnings("ignore")
from Orange.classification.utils.fasterrisk.utils import normalize_X, compute_logisticLoss_from_ExpyXB

class logRegModel:
    def __init__(self, X, y, lambda2=1e-8, intercept=True, original_lb=-5, original_ub=5):
        self.X = X
        self.X_normalized, self.X_mean, self.X_norm, self.scaled_feature_indices = normalize_X(self.X)
        self.n, self.p = self.X_normalized.shape
        self.y = y.reshape(-1).astype(float)
        self.yX = y.reshape(-1, 1) * self.X_normalized
        self.yXT = np.zeros((self.p, self.n))
        self.yXT[:] = np.transpose(self.yX)[:]
        self.beta0 = 0
        self.betas = np.zeros((self.p, ))
        self.ExpyXB = np.exp(self.y * self.beta0 + self.yX.dot(self.betas))

        self.intercept = intercept
        self.lambda2 = lambda2
        self.twoLambda2 = 2 * self.lambda2

        self.Lipschitz = 0.25 + self.twoLambda2
        self.lbs = original_lb * np.ones(self.p)
        self.lbs[self.scaled_feature_indices] *= self.X_norm[self.scaled_feature_indices]
        self.ubs = original_ub * np.ones(self.p)
        self.ubs[self.scaled_feature_indices] *= self.X_norm[self.scaled_feature_indices]

        self.total_child_added = 0
    
    def warm_start_from_original_beta0_betas(self, original_beta0, original_betas):
        # betas_initial has dimension (p+1, 1)
        self.original_beta0 = original_beta0
        self.original_betas = original_betas
        self.beta0, self.betas = self.transform_coefficients_to_normalized_space(self.original_beta0, self.original_betas)
        print("warmstart solution in normalized space is {} and {}".format(self.beta0, self.betas))
        self.ExpyXB = np.exp(self.y * self.beta0 + self.yX.dot(self.betas))

    def warm_start_from_beta0_betas(self, beta0, betas):
        self.beta0, self.betas = beta0, betas
        self.ExpyXB = np.exp(self.y * self.beta0 + self.yX.dot(self.betas))

    def warm_start_from_beta0_betas_ExpyXB(self, beta0, betas, ExpyXB):
        self.beta0, self.betas, self.ExpyXB = beta0, betas, ExpyXB

    def get_beta0_betas(self):
        return self.beta0, self.betas

    def get_beta0_betas_ExpyXB(self):
        return self.beta0, self.betas, self.ExpyXB
        
    def get_original_beta0_betas(self):
        return self.transform_coefficients_to_original_space(self.beta0, self.betas)

    def transform_coefficients_to_original_space(self, beta0, betas):
        original_betas = betas.copy()
        original_betas[self.scaled_feature_indices] = original_betas[self.scaled_feature_indices]/self.X_norm[self.scaled_feature_indices]
        original_beta0 = beta0 - np.dot(self.X_mean, original_betas)
        return original_beta0, original_betas

    def transform_coefficients_to_normalized_space(self, original_beta0, original_betas):
        betas = original_betas.copy()
        betas[self.scaled_feature_indices] = betas[self.scaled_feature_indices] * self.X_norm[self.scaled_feature_indices]
        beta0 = original_beta0 + self.X_mean.dot(original_betas)
        return beta0, betas

    def get_grad_at_coord(self, ExpyXB, betas_j, yX_j, j):
        # return -np.dot(1/(1+ExpyXB), self.yX[:, j]) + self.twoLambda2 * betas_j
        # return -np.inner(1/(1+ExpyXB), self.yX[:, j]) + self.twoLambda2 * betas_j
        # return -np.inner(np.reciprocal(1+ExpyXB), self.yX[:, j]) + self.twoLambda2 * betas_j
        return -np.inner(np.reciprocal(1+ExpyXB), yX_j) + self.twoLambda2 * betas_j
        # return -yX_j.dot(np.reciprocal(1+ExpyXB)) + self.twoLambda2 * betas_j

    def update_ExpyXB(self, ExpyXB, yX_j, diff_betas_j):
        ExpyXB *= np.exp(yX_j * diff_betas_j)

    def optimize_1step_at_coord(self, ExpyXB, betas, yX_j, j):
        # in-place modification, heck that ExpyXB and betas are passed by reference
        prev_betas_j = betas[j]
        current_betas_j = prev_betas_j
        grad_at_j = self.get_grad_at_coord(ExpyXB, current_betas_j, yX_j, j)
        step_at_j = grad_at_j / self.Lipschitz
        current_betas_j = prev_betas_j - step_at_j
        # current_betas_j = np.clip(current_betas_j, self.lbs[j], self.ubs[j])
        current_betas_j = max(self.lbs[j], min(self.ubs[j], current_betas_j))
        diff_betas_j = current_betas_j - prev_betas_j
        betas[j] = current_betas_j

        # ExpyXB *= np.exp(yX_j * diff_betas_j)
        self.update_ExpyXB(ExpyXB, yX_j, diff_betas_j)

    def finetune_on_current_support(self, ExpyXB, beta0, betas, total_CD_steps=100):

        support  = np.where(np.abs(betas) > 1e-9)[0]
        grad_on_support = -self.yXT[support].dot(np.reciprocal(1+ExpyXB)) + self.twoLambda2 * betas[support]
        abs_grad_on_support = np.abs(grad_on_support)
        support = support[np.argsort(-abs_grad_on_support)]

        loss_before = compute_logisticLoss_from_ExpyXB(ExpyXB) + self.lambda2 * betas[support].dot(betas[support])
        for steps in range(total_CD_steps): # number of iterations for coordinate descent

            if self.intercept:
                grad_intercept = -np.reciprocal(1+ExpyXB).dot(self.y)
                step_at_intercept = grad_intercept / (self.n * 0.25) # lipschitz constant is 0.25 at the intercept
                beta0 = beta0 - step_at_intercept
                ExpyXB *= np.exp(self.y * (-step_at_intercept))

            for j in support:
                self.optimize_1step_at_coord(ExpyXB, betas, self.yXT[j, :], j) # in-place modification on ExpyXB and betas
            
            if steps % 10 == 0:
                loss_after = compute_logisticLoss_from_ExpyXB(ExpyXB) + self.lambda2 * betas[support].dot(betas[support])
                if abs(loss_before - loss_after)/loss_after < 1e-8:
                    # print("break after {} steps; support size is {}".format(steps, len(support)))
                    break
                loss_before = loss_after
        
        return ExpyXB, beta0, betas
    
    def compute_yXB(self, beta0, betas):
        return self.y*(beta0 + np.dot(self.X_normalized, betas))
 


import numpy as np

from AnyQt import QtGui
from AnyQt.QtWidgets import (
    QTableWidget,
    QTableWidgetItem,
    QSlider,
    QLabel,
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
    QStyle,
    QToolTip,
    QStyleOptionSlider,
)
from AnyQt.QtCore import Qt, QRect
from AnyQt.QtGui import QPainter, QFontMetrics

from Orange.widgets import gui
from Orange.widgets.settings import ContextSetting
from Orange.widgets.widget import Input, Output, OWWidget, AttributeList, Msg
from Orange.data import Table
from Orange.classification import Model

from Orange.classification.scoringsheet import ScoringSheetModel
from Orange.classification.utils.fasterrisk.utils import (
    get_support_indices,
    get_all_product_booleans,
)


class ScoringSheetTable(QTableWidget):
    def __init__(self, main_widget, parent=None):
        """
        Initialize the ScoringSheetTable.

        It sets the column headers and connects the itemChanged
        signal to the handle_item_changed method.
        """
        super().__init__(parent)
        self.main_widget = main_widget
        self.setColumnCount(3)
        self.setHorizontalHeaderLabels(["Attribute Name", "Points", "Selected"])
        self.itemChanged.connect(self.handle_item_changed)

    def populate_table(self, attributes, coefficients):
        """
        Populates the table with the given attributes and coefficients.

        It creates a row for each attribute and populates the first two columns with
        the attribute name and coefficient respectively. The third column contains a
        checkbox that allows the user to select the attribute.
        """
        self.setRowCount(len(attributes))
        for i, (attr, coef) in enumerate(zip(attributes, coefficients)):
            # First column
            self.setItem(i, 0, QTableWidgetItem(attr))

            # Second column (align text to the right)
            coef_item = QTableWidgetItem(str(coef))
            coef_item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.setItem(i, 1, coef_item)

            # Third column (checkbox)
            checkbox = QTableWidgetItem()
            checkbox.setCheckState(Qt.Unchecked)
            self.setItem(i, 2, checkbox)

            for col in range(self.columnCount()):
                item = self.item(i, col)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable & ~Qt.ItemIsSelectable)

        # Resize columns to fit the contents
        self.resize_columns_to_contents()

    def resize_columns_to_contents(self):
        """
        Resize each column to fit the content.
        """
        for column in range(self.columnCount()):
            self.resizeColumnToContents(column)

    def handle_item_changed(self, item):
        """
        Handles the change in the state of the checkbox.

        It updates the slider value depending on the collected points.
        """
        if item.column() == 2:
            self.main_widget._update_slider_value()


class RiskSlider(QWidget):
    def __init__(self, points, probabilities, parent=None):
        super().__init__(parent)
        self.layout = QHBoxLayout(self)

        # Set the margins for the layout
        self.leftMargin = 20
        self.topMargin = 20
        self.rightMargin = 20
        self.bottomMargin = 20
        self.layout.setContentsMargins(
            self.leftMargin, self.topMargin, self.rightMargin, self.bottomMargin
        )

        # Setup the labels
        self.setup_labels()

        # Create the slider
        self.slider = QSlider(Qt.Horizontal, self)
        self.slider.setEnabled(False)
        self.layout.addWidget(self.slider)

        self.points = points
        self.probabilities = probabilities
        self.setup_slider()

        # Set the margin for drawing text
        self.textMargin = 1

        # This is needed to show the tooltip when the mouse is over the slider thumb
        self.slider.installEventFilter(self)
        self.setMouseTracking(True)
        self.target_class = None

        self.label_frequency = 1

    def setup_labels(self):
        """
        Set up the labels for the slider.

        It creates a vertical layout for the labels and adds it to the main layout.
        It is only called once when the widget is initialized.
        """
        # Create the labels for the slider
        self.label_layout = QVBoxLayout()
        # Add the label for the points "Points:"
        self.points_label = QLabel("<b>Total:</b>")
        self.label_layout.addWidget(self.points_label)
        # Add stretch to the label layout
        self.label_layout.addSpacing(23)
        # Add the label for the probability "Probability:"
        self.probability_label = QLabel("<b>Probabilities (%):</b>")
        self.label_layout.addWidget(self.probability_label)
        self.layout.addLayout(self.label_layout)
        # Add a spacer
        self.layout.addSpacing(28)

    def setup_slider(self):
        """
        Set up the slider with the given points and probabilities.

        It sets the minimum and maximum values (of the indexes for the ticks) of the slider.
        It is called when the points and probabilities are updated.
        """
        self.slider.setMinimum(0)
        self.slider.setMaximum(len(self.points) - 1 if self.points else 0)
        self.slider.setTickPosition(QSlider.TicksBothSides)
        self.slider.setTickInterval(1)  # Set tick interval

    def move_to_value(self, value):
        """
        Move the slider to the closest tick mark to the given value.
        """
        if not self.points:
            return
        closest_point_index = min(
            range(len(self.points)), key=lambda i: abs(self.points[i] - value)
        )
        self.slider.setValue(closest_point_index)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.update_label_frequency()
        self.update()

    def update_label_frequency(self):
        """
        Update the label frequency based on the width of the slider and the number of points.

        Label frequency determines how many labels are shown on the slider.
        """
        total_width = self.slider.width()
        label_width = QFontMetrics(self.font()).boundingRect("100.0%").width()
        max_labels = total_width // label_width

        frequencies = [1, 2, 5, 10, 20, 50, 100]
        for frequency in frequencies:
            if max_labels >= len(self.points) / frequency:
                self.label_frequency = frequency
                break

    def paintEvent(self, event):
        """
        Paint the point and probabilitie labels above and below the tick marks respectively.
        """
        super().paintEvent(event)

        if not self.points:
            return

        painter = QPainter(self)
        fm = QFontMetrics(painter.font())

        for i, point in enumerate(self.points):
            if i % self.label_frequency == 0:
                # Calculate the x position of the tick mark
                x_pos = (
                    QStyle.sliderPositionFromValue(
                        self.slider.minimum(),
                        self.slider.maximum(),
                        i,
                        self.slider.width(),
                    )
                    + self.slider.x()
                )

                # Draw the point label above the tick mark
                point_str = str(point)
                point_rect = fm.boundingRect(point_str)
                point_x = int(x_pos - point_rect.width() / 2)
                point_y = int(self.slider.y() - self.textMargin - point_rect.height())
                painter.drawText(
                    QRect(point_x, point_y, point_rect.width(), point_rect.height()),
                    Qt.AlignCenter,
                    point_str,
                )

                # Draw the probability label below the tick mark
                prob_str = str(round(self.probabilities[i], 1)) + "%"
                prob_rect = fm.boundingRect(prob_str)
                prob_x = int(x_pos - prob_rect.width() / 2)
                prob_y = int(self.slider.y() + self.slider.height() + self.textMargin)
                painter.drawText(
                    QRect(prob_x, prob_y, prob_rect.width(), prob_rect.height()),
                    Qt.AlignCenter,
                    prob_str,
                )

        painter.end()

    def eventFilter(self, watched, event):
        """
        Event filter to intercept hover events on the slider.

        This is needed to show the tooltip when the mouse is over the slider thumb.
        """
        if watched == self.slider and isinstance(event, QtGui.QHoverEvent):
            # Handle the hover event when it's over the slider
            self.handle_hover_event(event.pos())
            return True
        else:
            # Call the base class method to continue default event processing
            return super().eventFilter(watched, event)

    def handle_hover_event(self, pos):
        """
        Handle hover events for the slider.

        Display the tooltip when the mouse is over the slider thumb.
        """
        thumbRect = self.get_thumb_rect()
        if thumbRect.contains(pos) and self.points:
            value = self.slider.value()
            points = self.points[value]
            probability = self.probabilities[value]
            tooltip = str(
                f"<b>{self.target_class}</b>\n "
                f"<hr style='margin: 0px; padding: 0px; border: 0px; height: 1px; background-color: #000000'>"
                f"<b>Points:</b> {int(points)}<br>"
                f"<b>Probability:</b> {probability:.1f}%"
            )
            QToolTip.showText(self.slider.mapToGlobal(pos), tooltip)
        else:
            QToolTip.hideText()

    def get_thumb_rect(self):
        """
        Get the rectangle of the slider thumb.
        """
        opt = QStyleOptionSlider()
        self.slider.initStyleOption(opt)

        style = self.slider.style()

        # Get the area of the slider that contains the handle
        handle_rect = style.subControlRect(
            QStyle.CC_Slider, opt, QStyle.SC_SliderHandle, self.slider
        )

        # Calculate the position and size of the thumb
        thumb_x = handle_rect.x()
        thumb_y = handle_rect.y()
        thumb_width = handle_rect.width()
        thumb_height = handle_rect.height()

        return QRect(thumb_x, thumb_y, thumb_width, thumb_height)


class OWScoringSheetViewer(OWWidget):
    """
    Allows visualization of the scoring sheet model.
    """

    name = "Scoring Sheet Viewer"
    description = "Visualize the scoring sheet model."
    want_control_area = False
    icon = "icons/ScoringSheetViewer.svg"
    replaces = [
        "orangecontrib.prototypes.widgets.owscoringsheetviewer.OWScoringSheetViewer"
    ]
    priority = 2010
    keywords = "scoring sheet viewer"

    class Inputs:
        classifier = Input("Classifier", Model)
        data = Input("Data", Table)

    class Outputs:
        features = Output("Features", AttributeList)

    target_class_index = ContextSetting(0)

    class Error(OWWidget.Error):
        invalid_classifier = Msg(
            "Scoring Sheet Viewer only accepts a Scoring Sheet model."
        )

    class Information(OWWidget.Information):
        multiple_instances = Msg(
            "The input data contains multiple instances. Only the first instance will be used."
        )

    def __init__(self):
        super().__init__()
        self.data = None
        self.instance = None
        self.instance_points = []
        self.classifier = None
        self.coefficients = None
        self.attributes = None
        self.all_scores = None
        self.all_risks = None
        self.domain = None
        self.old_target_class_index = self.target_class_index

        self._setup_gui()
        self.resize(700, 400)

    # GUI Methods ----------------------------------------------------------------------------------

    def _setup_gui(self):
        # Create a new widget box for the combo box in the main area
        combo_box_layout = gui.widgetBox(self.mainArea, orientation="horizontal")
        self.class_combo = gui.comboBox(
            combo_box_layout,
            self,
            "target_class_index",
            callback=self._class_combo_changed,
        )
        self.class_combo.setFixedWidth(100)
        combo_box_layout.layout().addWidget(QLabel("Target class:"))
        combo_box_layout.layout().addWidget(self.class_combo)
        combo_box_layout.layout().addStretch()

        self.coefficient_table = ScoringSheetTable(main_widget=self, parent=self)
        gui.widgetBox(self.mainArea).layout().addWidget(self.coefficient_table)

        self.risk_slider = RiskSlider([], [], self)
        gui.widgetBox(self.mainArea).layout().addWidget(self.risk_slider)

    def _reset_ui_to_original_state(self):
        """
        Reset all UI components to their original state.
        """
        # Reset the coefficient table
        self.coefficient_table.clearContents()
        self.coefficient_table.setRowCount(0)

        # Reset the risk slider
        self.risk_slider.slider.setValue(0)
        self.risk_slider.points = []
        self.risk_slider.probabilities = []
        self.risk_slider.setup_slider()
        self.risk_slider.update()

        # Reset class combo box
        self.class_combo.clear()

    def _populate_interface(self):
        """Populate the scoring sheet based on extracted data."""
        if self.attributes and self.coefficients:
            self.coefficient_table.populate_table(self.attributes, self.coefficients)

            # Update points and probabilities in the custom slider
            class_var_name = self.domain.class_vars[0].name
            class_var_value = self.domain.class_vars[0].values[self.target_class_index]

            self.risk_slider.points = self.all_scores
            self.risk_slider.probabilities = self.all_risks
            self.risk_slider.target_class = f"{class_var_name} = {class_var_value}"
            self.risk_slider.setup_slider()
            self.risk_slider.update()

    def _update_slider_value(self):
        """
        Updates the slider value to reflect the total points collected.

        This method is called when user changes the state of the checkbox in the coefficient table.
        """
        if not self.coefficient_table:
            return
        total_coefficient = sum(
            float(self.coefficient_table.item(row, 1).text())
            for row in range(self.coefficient_table.rowCount())
            if self.coefficient_table.item(row, 2)
            and self.coefficient_table.item(row, 2).checkState() == Qt.Checked
        )
        self.risk_slider.move_to_value(total_coefficient)

    def _update_controls(self):
        """
        It updates the interface components based on the extracted data.

        This method is called when the user inputs data, changes the classifier or the target class.
        """
        self._populate_interface()
        self._update_slider_value()
        self._setup_class_combo()
        self._set_instance_points()

    # Class Combo Methods --------------------------------------------------------------------------

    def _setup_class_combo(self):
        """
        This method is used to populate the class combo box with the target classes.
        """
        self.class_combo.clear()
        if self.domain is not None:
            values = self.domain.class_vars[0].values
            if values:
                self.class_combo.addItems(values)
                self.class_combo.setCurrentIndex(self.target_class_index)

    def _class_combo_changed(self):
        """
        This method is called when the user changes the target class.
        It updates the interface components based on the selected class.
        """
        self.target_class_index = self.class_combo.currentIndex()
        if self.target_class_index == self.old_target_class_index:
            return
        self.old_target_class_index = self.target_class_index

        self._adjust_for_target_class()
        self._update_controls()

    def _adjust_for_target_class(self):
        """
        Adjusts the coefficients, scores, and risks for the negative/positive class.

        This allows user to select the target class and see the
        corresponding coefficients, scores, and risks.
        """
        # Negate the coefficients
        self.coefficients = [-coef for coef in self.coefficients]
        # Negate the scores
        self.all_scores = [-score if score != 0 else score for score in self.all_scores]
        self.all_scores.sort()
        # Adjust the risks
        self.all_risks = [100 - risk for risk in self.all_risks]
        self.all_risks.sort()

    # Classifier Input Methods ---------------------------------------------------------------------

    def _extract_data_from_model(self, classifier):
        """
        Extracts the attributes, non-zero coefficients, all possible
        scores, and corresponding probabilities from the model.
        """
        model = classifier.model

        # 1. Extracting attributes and non-zero coefficients
        nonzero_indices = get_support_indices(model.coefficients)
        attributes = [model.featureNames[i] for i in nonzero_indices]
        coefficients = [int(model.coefficients[i]) for i in nonzero_indices]

        # 2. Extracting possible points and corresponding probabilities
        len_nonzero_indices = len(nonzero_indices)
        # If we have less than 10 attributes, we can calculate all possible combinations of scores.
        if len_nonzero_indices <= 10:
            all_product_booleans = get_all_product_booleans(len_nonzero_indices)
            all_scores = all_product_booleans.dot(model.coefficients[nonzero_indices])
            all_scores = np.unique(all_scores)
        # If there are more than 10 non-zero coefficients, calculating all possible combinations
        # of scores might be computationally intensive. Instead, the method calculates all possible
        # scores from the training dataset (X_train) and then picks some quantile points
        # (in this case, a maximum of 20) to represent the possible scores.
        else:
            all_scores = model.X_train.dot(model.coefficients)
            all_scores = np.unique(all_scores)
            quantile_len = min(20, len(all_scores))
            quantile_points = np.asarray(range(1, 1 + quantile_len)) / quantile_len
            all_scores = np.quantile(
                all_scores, quantile_points, method="closest_observation"
            )

        all_scaled_scores = (model.intercept + all_scores) / model.multiplier
        all_risks = 1 / (1 + np.exp(-all_scaled_scores))

        self.attributes = attributes
        self.coefficients = coefficients
        self.all_scores = all_scores.tolist()
        self.all_risks = (all_risks * 100).tolist()
        self.domain = classifier.domain

        # For some reason when leading the model the scores and probabilities are
        # set for the wrong target class. This is a workaround to fix that.
        self._adjust_for_target_class()

    def _is_valid_classifier(self, classifier):
        """Check if the classifier is a valid ScoringSheetModel."""
        if not isinstance(classifier, ScoringSheetModel):
            self.Error.invalid_classifier()
            return False
        return True

    def _clear_classifier_data(self):
        """Clear classifier data and associated interface components."""
        self.coefficients = None
        self.attributes = None
        self.all_scores = None
        self.all_risks = None
        self.classifier = None
        self.Outputs.features.send(None)

    # Data Input Methods ---------------------------------------------------------------------------

    def _clear_table_data(self):
        """Clear data and associated interface components."""
        self.data = None
        self.instance = None
        self.instance_points = []
        self._set_table_checkboxes()

    def _set_instance_points(self):
        """
        Initializes the instance and its points and sets the checkboxes in the coefficient table.
        """
        if self.data and self.domain is not None:
            self._init_instance_points()

        self._set_table_checkboxes()

    def _set_table_checkboxes(self):
        """
        Sets the checkboxes in the coefficient table based on the instance points.
        Or clears the checkboxes if the instance points are not initialized.
        """
        for row in range(self.coefficient_table.rowCount()):
            if self.instance_points and self.instance_points[row] != 0:
                self.coefficient_table.item(row, 2).setCheckState(Qt.Checked)
            else:
                self.coefficient_table.item(row, 2).setCheckState(Qt.Unchecked)

    def _init_instance_points(self):
        """
        Initialize the instance which is used to show the points collected for each attribute.
        Get the values of the features for the instance and store them in a list.
        """
        instances = self.data.transform(self.domain)
        self.instance = instances[0]
        self.instance_points = [
            self.instance.list[i]
            for i in get_support_indices(self.classifier.model.coefficients)
        ]

    # Input Methods --------------------------------------------------------------------------------

    @Inputs.classifier
    def set_classifier(self, classifier):
        self.Error.invalid_classifier.clear()
        if not classifier or not self._is_valid_classifier(classifier):
            self._clear_classifier_data()
            self._reset_ui_to_original_state()
            return

        self.classifier = classifier
        self._extract_data_from_model(classifier)
        self._update_controls()
        # Output the features
        self.Outputs.features.send(
            AttributeList(
                [feature for feature in self.domain if feature.name in self.attributes]
            )
        )

    @Inputs.data
    def set_data(self, data):
        self.Information.multiple_instances.clear()
        if not data or len(data) < 1:
            self._clear_table_data()
            return

        self.data = data
        if len(data) > 1:
            self.Information.multiple_instances()
        self._update_controls()


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from Orange.classification.scoringsheet import ScoringSheetLearner

    mock_data = Table("heart_disease")
    mock_learner = ScoringSheetLearner(15, 5, 5, None)
    mock_model = mock_learner(mock_data)
    WidgetPreview(OWScoringSheetViewer).run(
        set_classifier=mock_model, set_data=mock_data
    )



"""
======
Models
======

Classifiers and regressors.

"""

NAME = "Model"

ID = "orange.widgets.model"

DESCRIPTION = "Prediction"

BACKGROUND = "#FAC1D9"

ICON = "icons/Category-Model.svg"

PRIORITY = 4



# This module mimics unittest's default main with --verbose, but reduces
# the output by not showing successful tests.
#
# stdout/stderr are redirected into the same stream (stdout). The name of the
# test and its description (if any) are printed only if the test fails or the
# tests causes some printout. For this, the module defines a stream with
# 'preambule' that is, a string that is printed before the next print.
# The preambule is set at the beginning at each test function.

import os
import sys
from io import IOBase
from unittest import TestResult, TextTestRunner
from unittest.main import main

__unittest = True

if os.getenv("CI"):
    # Remove the current path to prevent tox from importing sources instead of
    # installed files (with compiled binaries)
    sys.path.remove(os.path.split(__file__)[0])


class PreambuleStream(IOBase):
    def __init__(self):
        super().__init__()
        self.preambule = ""
        self.line_before_msg = False

    def set_preambule(self, preambule):
        self.preambule = preambule

    def writable(self):
        return True

    def writelines(self, lines):
        return self.write("".join(lines))

    def write(self, s):
        if self.preambule:
            _stdout.write("\n" + self.preambule + "\n")
            self.preambule = ""
        self.line_before_msg = True
        return _stdout.write(s)

    def write_msg(self, s):
        if self.line_before_msg:
            _stdout.write("\n")
        _stdout.write(self.preambule + " ... " + s)
        self.line_before_msg = False
        self.preambule = ""

    def flush(self):
        _stdout.flush()


class QuietTestResult(TestResult):
    separator1 = '=' * 70
    separator2 = '-' * 70

    def startTest(self, test):
        super().startTest(test)
        sys.stdout.set_preambule(self.getDescription(test))

    def stopTest(self, test):
        super().stopTest(test)
        sys.stdout.set_preambule("")

    @staticmethod
    def getDescription(test):
        doc_first_line = test.shortDescription()
        if doc_first_line:
            return '\n'.join((str(test), doc_first_line))
        else:
            return str(test)

    def addError(self, test, err):
        super().addError(test, err)
        sys.stdout.write_msg("ERROR\n")

    def addFailure(self, test, err):
        super().addError(test, err)
        sys.stdout.write_msg("FAIL\n")

    def printErrors(self):
        sys.stdout.set_preambule("")
        print()
        self.printErrorList('ERROR', self.errors)
        self.printErrorList('FAIL', self.failures)

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            print(self.separator1)
            print("%s: %s" % (flavour,self.getDescription(test)))
            print(self.separator2)
            print("%s" % err)


_stdout = sys.stdout
sys.stderr = sys.stdout = PreambuleStream()

testRunner = TextTestRunner(
    resultclass=QuietTestResult,
    stream=sys.stdout,
    verbosity=2)

main(module=None, testRunner=testRunner)



from setuptools import setup

setup(
    name="Demo",
    packages=["orangedemo"],
    package_data={"orangedemo": ["icons/*.svg"]},
    classifiers=["Example :: Invalid"],
    # Declare orangedemo package to contain widgets for the "Demo" category
    entry_points={"orange.widgets": "Demo = orangedemo"},
)



from unittest.mock import patch, MagicMock

import numpy as np

from Orange.data import Domain, Table, ContinuousVariable
from Orange.preprocess import Normalize, SklImpute

from .base import Benchmark, benchmark


class SetUpData:

    def setUp(self):
        cols = 1000
        rows = 1000
        cont = [ContinuousVariable(str(i)) for i in range(cols)]
        self.domain = Domain(cont)
        self.single = Domain([ContinuousVariable("0")])
        self.table = Table.from_numpy(
            self.domain,
            np.random.RandomState(0).randint(0, 2, (rows, len(self.domain.variables))))
        self.normalized_domain = Normalize()(self.table).domain


class BenchNormalize(SetUpData, Benchmark):

    @benchmark(number=5)
    def bench_normalize_only_transform(self):
        self.table.transform(self.normalized_domain)

    @benchmark(number=5)
    def bench_normalize_only_parameters(self):
        # avoid benchmarking transformation
        with patch("Orange.data.Table.transform", MagicMock()):
            Normalize()(self.table)


class BenchSklImpute(SetUpData, Benchmark):

    @benchmark(number=5)
    def bench_sklimpute(self):
        SklImpute()(self.table)



from AnyQt.QtCore import Qt

from orangewidget.tests.base import GuiTest

import Orange
from Orange.data import Table
from Orange.widgets.data.owtable import RichTableModel, TableBarItemDelegate, \
    TableDataDelegate
from Orange.widgets.utils.distmatrixmodel import DistMatrixModel
from Orange.widgets.utils.tableview import TableView

from .base import benchmark, Benchmark


class BaseBenchTableView(GuiTest, Benchmark):
    def setUp(self) -> None:
        super().setUp()
        self.view = TableView()
        self.view.resize(1024, 760)

    def tearDown(self) -> None:
        super().tearDown()
        self.view.setModel(None)
        del self.view


class BaseBenchTableModel(BaseBenchTableView):
    def setUp(self) -> None:
        super().setUp()
        data = Table("brown-selected")
        self.model = RichTableModel(data)
        self.view.setModel(self.model)


class BenchTableView(BaseBenchTableView):
    def setUp(self) -> None:
        super().setUp()
        data = Table("brown-selected")
        self.delegate = TableDataDelegate(
            self.view, roles=(
                Qt.DisplayRole, Qt.BackgroundRole, Qt.TextAlignmentRole
            )
        )
        self.view.setItemDelegate(self.delegate)
        self.model = RichTableModel(data)
        self.view.setModel(self.model)
        self.option = self.view.viewOptions()
        self.index_00 = self.model.index(0, 0)

    def tearDown(self) -> None:
        super().tearDown()
        del self.model
        del self.delegate
        del self.option
        del self.index_00

    @benchmark(number=10, warmup=1, repeat=3)
    def bench_paint(self):
        _ = self.view.grab()

    @benchmark(number=10, warmup=1, repeat=3)
    def bench_init_style_option(self):
        self.delegate.initStyleOption(self.option, self.index_00)


class BenchTableBarItemDelegate(BenchTableView):
    def setUp(self) -> None:
        super().setUp()
        self.delegate = TableBarItemDelegate(self.view)
        # self.delegate = gui.TableBarItem()
        self.view.setItemDelegate(self.delegate)


class BenchDistanceDelegate(BaseBenchTableView):
    def setUp(self) -> None:
        super().setUp()
        data = Table("iris")
        dist = Orange.distance.Euclidean(data)
        self.model = DistMatrixModel()
        self.model.set_data(dist)
        self.view.setModel(self.model)

    def tearDown(self) -> None:
        super().tearDown()
        del self.model

    @benchmark(number=3, warmup=1, repeat=10)
    def bench_paint(self):
        self.view.grab()



from functools import lru_cache
from typing import Callable, Dict, List, Tuple, Union, Type

import pandas as pd

from Orange.data import Domain, Table, Variable, table_from_frame, table_to_frame
from Orange.util import dummy_callback


class OrangeTableGroupBy:
    """
    A class representing the result of the groupby operation on Orange's
    Table and offers aggregation functionality on groupby object. It wraps
    Panda's GroupBy object.

    Attributes
    ----------
    table
        Table to be grouped
    by
        Variable used for grouping. Resulting groups are defined with unique
        combinations of those values.

    Examples
    --------
    from Orange.data import Table

    table = Table("iris")
    gb = table.groupby([table.domain["iris"]])
    aggregated_table = gb.aggregate(
        {table.domain["sepal length"]: ["mean", "median"],
         table.domain["petal length"]: ["mean"]}
    )
    """

    def __init__(self, table: Table, by: List[Variable]):
        self.table = table

        df = table_to_frame(table, include_metas=True)
        # observed=True keeps only groups with at leas one instance
        self.group_by = df.groupby([a.name for a in by], observed=True)
        self.by = tuple(by)

        # lru_cache that is caches on the object level
        self.compute_aggregation = lru_cache()(self._compute_aggregation)

    AggDescType = Union[str,
                    Callable,
                    Tuple[str, Union[str, Callable]],
                    Tuple[str, Union[str, Callable], Union[Type[Variable], bool]]
    ]

    def aggregate(
        self,
        aggregations: Dict[Variable, List[AggDescType]],
        callback: Callable = dummy_callback,
    ) -> Table:
        """
        Compute aggregations for each group

        Parameters
        ----------
        aggregations
            The dictionary that defines aggregations that need to be computed
            for variables. We support three formats:
            - {variable name: [agg function 1, agg function 2]}
            - {variable name: [(agg name 1, agg function 1),  (agg name 1, agg function 1)]}
            - {variable name: [(agg name 1, agg function 1, output_variable_type1), ...]}
            Where agg name is the aggregation name used in the output column name.
            Aggregation function can be either function or string that defines
            aggregation in Pandas (e.g. mean).
            output_variable_type can be a type for a new variable, True to copy
            the input variable, or False to create a new variable of the same type
            as the input
        callback
            Callback function to report the progress

        Returns
        -------
        Table that includes aggregation columns. Variables that are used for
        grouping are in metas.
        """
        num_aggs = sum(len(aggs) for aggs in aggregations.values())
        count = 0

        result_agg = []
        output_variables = []
        for col, aggs in aggregations.items():
            for agg in aggs:
                res, var = self._compute_aggregation(col, agg)
                result_agg.append(res)
                output_variables.append(var)
                count += 1
                callback(count / num_aggs * 0.8)

        agg_table = self._aggregations_to_table(result_agg, output_variables)
        callback(1)
        return agg_table

    def _compute_aggregation(
            self, col: Variable, agg: AggDescType) -> Tuple[pd.Series, Variable]:
        # use named aggregation to avoid issues with same column names when reset_index
        if isinstance(agg, tuple):
            name, agg, var_type, *_ = (*agg, None)
        else:
            name = agg if isinstance(agg, str) else agg.__name__
            var_type = None
        col_name = f"{col.name} - {name}"
        agg_col = self.group_by[col.name].agg(**{col_name: agg})
        if col.is_discrete and var_type is True:
            dtype = pd.CategoricalDtype(categories=col.values, ordered=True)
            agg_col = agg_col.astype(dtype)
        if var_type is True:
            var = col.copy(name=col_name)
        elif var_type is False:
            var = col.make(name=col_name)
        elif var_type is None:
            var = None
        else:
            assert issubclass(var_type, Variable)
            var = var_type.make(name=col_name)
        return agg_col, var

    def _aggregations_to_table(
            self,
            aggregations: List[pd.Series],
            output_variables: List[Union[Variable, None]]) -> Table:
        """Concatenate aggregation series and convert back to Table"""
        if aggregations:
            df = pd.concat(aggregations, axis=1)
        else:
            # when no aggregation is computed return a table with gropby columns
            df = self.group_by.first()
            df = df.drop(columns=df.columns)
        gb_attributes = df.index.names
        df = df.reset_index()  # move group by var that are in index to columns
        table = table_from_frame(df, variables=(*self.by, *output_variables))

        # group by variables should be last two columns in metas in the output
        metas = table.domain.metas
        new_metas = [m for m in metas if m.name not in gb_attributes] + [
            table.domain[n] for n in gb_attributes
        ]
        new_domain = Domain(
            [var for var in table.domain.attributes if var.name not in gb_attributes],
            metas=new_metas,
        )
        # keeps input table's type - e.g. output is Corpus if input Corpus
        return self.table.from_table(new_domain, table)



# pylint: disable=missing-docstring
# pylint: disable=protected-access
import unittest
from unittest.mock import Mock, patch
from packaging.version import parse as parse_version

import numpy as np
from sklearn import __version__ as sklearn_version
from AnyQt.QtCore import Qt
from AnyQt.QtTest import QTest
from AnyQt.QtWidgets import QApplication
import baycomp

from Orange.classification import MajorityLearner, LogisticRegressionLearner, \
    RandomForestLearner
from Orange.classification.majority import ConstantModel
from Orange.data import Table, Domain, DiscreteVariable, ContinuousVariable
from Orange.evaluation import Results, TestOnTestData, scoring
from Orange.evaluation.scoring import ClassificationScore, RegressionScore, \
    Score
from Orange.base import Learner, Model
from Orange.modelling import ConstantLearner
from Orange.regression import MeanLearner, PLSRegressionLearner
from Orange.widgets.evaluate.owtestandscore import (
    OWTestAndScore, results_one_vs_rest)
from Orange.widgets.settings import (
    ClassValuesContextHandler, PerfectDomainContextHandler)
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate, possible_duplicate_table
from Orange.tests import test_filename


class BadLearner(Learner):
    def fit(self, *_, **_2):  # pylint: disable=arguments-differ
        return 1 / 0


class TestOWTestAndScore(WidgetTest):
    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(OWTestAndScore)  # type: OWTestAndScore

        self.scores_domain = Domain(
            [ContinuousVariable("a"), ContinuousVariable("b")],
            [DiscreteVariable("c", values=("y", "n"))])

        self.scores_table_values = [[1, 1, 1.23, 23.8], [1., 2., 3., 4.]]

    def tearDown(self):
        self.widget.onDeleteWidget()
        super().tearDown()

    def test_basic(self):
        data = Table("iris")[::15]
        self.send_signal(self.widget.Inputs.train_data, data)
        self.send_signal(self.widget.Inputs.learner, MajorityLearner(), 0)
        res = self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsInstance(res, Results)
        self.assertIsNotNone(res.domain)
        self.assertIsNotNone(res.data)
        self.assertIsNotNone(res.probabilities)

        self.send_signal(self.widget.Inputs.learner, None, 0)
        res = self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsNone(res)

        data = Table("housing")[::10]
        self.send_signal(self.widget.Inputs.train_data, data)
        self.send_signal(self.widget.Inputs.learner, MeanLearner(), 0)
        res = self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsInstance(res, Results)
        self.assertIsNotNone(res.domain)
        self.assertIsNotNone(res.data)

    def test_multiple_learners(self):
        def check_evres_names(expeced):
            res = self.get_output(self.widget.Outputs.evaluations_results)
            self.assertSequenceEqual(res.learner_names, expeced)

        data = Table("iris")[::15].copy()
        m1 = MajorityLearner()
        m1.name = "M1"
        m2 = MajorityLearner()
        m2.name = "M2"
        self.send_signal(self.widget.Inputs.train_data, data)
        self.send_signal(self.widget.Inputs.learner, m1, 1)
        self.send_signal(self.widget.Inputs.learner, m2, 2)
        res = self.get_output(self.widget.Outputs.evaluations_results)
        np.testing.assert_equal(res.probabilities[0], res.probabilities[1])
        check_evres_names(["M1", "M2"])
        self.send_signal(self.widget.Inputs.learner, None, 1)
        check_evres_names(["M2"])
        self.send_signal(self.widget.Inputs.learner, m1, 1)
        check_evres_names(["M1", "M2"])
        self.send_signal(self.widget.Inputs.learner,
                         self.widget.Inputs.learner.closing_sentinel, 1)
        check_evres_names(["M2"])
        self.send_signal(self.widget.Inputs.learner, m1, 1)
        check_evres_names(["M2", "M1"])

    def test_testOnTest(self):
        data = Table("iris")
        self.send_signal(self.widget.Inputs.train_data, data)
        self.widget.resampling = OWTestAndScore.TestOnTest
        self.send_signal(self.widget.Inputs.test_data, data)

    def test_testOnTest_incompatible_domain(self):
        iris = Table("iris")
        self.send_signal(self.widget.Inputs.train_data, iris)
        self.send_signal(self.widget.Inputs.learner, LogisticRegressionLearner(), 0)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertFalse(self.widget.Error.test_data_incompatible.is_shown())
        self.widget.resampling = OWTestAndScore.TestOnTest
        # test data with the same class (otherwise the widget shows a different error)
        # and a non-nan X
        iris_test = iris.transform(
            Domain([ContinuousVariable("x")], class_vars=iris.domain.class_vars)).copy()
        with iris_test.unlocked():
            iris_test.X[:, 0] = 1
        self.send_signal(self.widget.Inputs.test_data, iris_test)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertTrue(self.widget.Error.test_data_incompatible.is_shown())

    def test_CrossValidationByFeature(self):
        data = Table("iris")
        attrs = data.domain.attributes
        domain = Domain(attrs[:-1], attrs[-1], data.domain.class_vars)
        data_with_disc_metas = Table.from_table(domain, data)
        rb = self.widget.controls.resampling.buttons[OWTestAndScore.FeatureFold]

        self.send_signal(self.widget.Inputs.learner, ConstantLearner(), 0)
        self.send_signal(self.widget.Inputs.train_data, data)
        self.assertFalse(rb.isEnabled())
        self.assertFalse(self.widget.features_combo.isEnabled())
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)

        self.send_signal(self.widget.Inputs.train_data, data_with_disc_metas)
        self.assertTrue(rb.isEnabled())
        rb.click()
        self.assertEqual(self.widget.resampling, OWTestAndScore.FeatureFold)
        self.assertTrue(self.widget.features_combo.isEnabled())
        self.assertEqual(self.widget.features_combo.currentText(), "iris")
        self.assertEqual(len(self.widget.features_combo.model()), 1)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)

        self.send_signal(self.widget.Inputs.train_data, None)
        self.assertFalse(rb.isEnabled())
        self.assertEqual(self.widget.resampling, OWTestAndScore.KFold)
        self.assertFalse(self.widget.features_combo.isEnabled())

    def test_migrate_removes_invalid_contexts(self):
        context_invalid = ClassValuesContextHandler().new_context([0, 1, 2])
        context_valid = PerfectDomainContextHandler().new_context(*[[]] * 4)
        settings = {'context_settings': [context_invalid, context_valid]}
        self.widget.migrate_settings(settings, 2)
        self.assertEqual(settings['context_settings'], [context_valid])

    def test_migrate_shown_scores(self):
        settings = {"score_table": {"shown_scores": {"Sensitivity"}}}
        self.widget.migrate_settings(settings, 3)
        self.assertTrue(settings["score_table"]["show_score_hints"]["Sensitivity"])

    def test_memory_error(self):
        """
        Handling memory error.
        GH-2316
        """
        data = Table("iris")[::15]
        self.send_signal(self.widget.Inputs.train_data, data)
        self.assertFalse(self.widget.Error.memory_error.is_shown())

        with unittest.mock.patch(
                "Orange.evaluation.testing.Results.get_augmented_data",
                side_effect=MemoryError):
            self.send_signal(self.widget.Inputs.learner, MajorityLearner(), 0, wait=5000)
            self.assertTrue(self.widget.Error.memory_error.is_shown())

    def test_one_class_value(self):
        """
        Data with a class with one value causes widget to crash when that value
        is selected.
        GH-2351
        """
        table = Table.from_list(
            Domain(
                [ContinuousVariable("a"), ContinuousVariable("b")],
                [DiscreteVariable("c", values=("y",))]),
            list(zip(
                [42.48, 16.84, 15.23, 23.8],
                [1., 2., 3., 4.],
                "yyyy"))
        )
        self.widget.n_folds = 0
        self.assertFalse(self.widget.Error.train_data_error.is_shown())
        self.send_signal(self.widget.Inputs.train_data, table)
        self.send_signal(self.widget.Inputs.learner, MajorityLearner(), 0, wait=1000)
        self.assertTrue(self.widget.Error.train_data_error.is_shown())

    def test_data_errors(self):
        """ Test all data_errors """

        def assertErrorShown(data, is_shown, message):
            self.send_signal(self.widget.Inputs.train_data, data)
            self.assertEqual(is_shown, self.widget.Error.train_data_error.is_shown())
            self.assertEqual(message, str(self.widget.Error.train_data_error))

        data = Table("iris")[::30].copy()
        with data.unlocked():
            data.Y[:] = np.nan

        iris_empty_x = Table.from_table(
            Domain([], data.domain.class_var), Table("iris")
        )

        for data, is_shown, message in zip(
                [None, data, Table("iris")[:30], iris_empty_x, data[:0]],
                [False, True, True, True, True],
                [
                    "",
                    "Target variable has no values.",
                    "Target variable has only one value.",
                    "Data has no features to learn from.",
                    "Train dataset is empty."
                ]

        ):
            assertErrorShown(data, is_shown, message)

    def test_addon_scorers(self):
        try:
            # These classes are registered, pylint: disable=unused-variable
            class NewScore(Score):
                class_types = (DiscreteVariable, ContinuousVariable)
                name = "new scorer"

                @staticmethod
                def is_compatible(domain: Domain) -> bool:
                    return True

            class NewClassificationScore(ClassificationScore):
                name = "new classification scorer"
                default_visible = False

            class NewRegressionScore(RegressionScore):
                pass

            widget = self.create_widget(OWTestAndScore)
            header = widget.score_table.view.horizontalHeader()
            self.send_signal(widget.Inputs.train_data, Table("iris"))
            scorer_names = [scorer.name for scorer in widget.scorers]
            self.assertIn("new scorer", scorer_names)
            self.assertFalse(header.isSectionHidden(3 + scorer_names.index("new scorer")))
            self.assertIn("new classification scorer", scorer_names)
            self.assertTrue(header.isSectionHidden(3 + scorer_names.index("new classification scorer")))
            self.assertNotIn("NewRegressionScore", scorer_names)
            model = widget.score_table.model


            self.send_signal(widget.Inputs.train_data, Table("housing"))
            scorer_names = [scorer.name for scorer in widget.scorers]
            self.assertIn("new scorer", scorer_names)
            self.assertNotIn("new classification scorer", scorer_names)
            self.assertIn("NewRegressionScore", scorer_names)

            self.send_signal(widget.Inputs.train_data, None)
            self.assertEqual(widget.scorers, [])
        finally:
            del Score.registry["NewScore"]  # pylint: disable=no-member
            del Score.registry["NewClassificationScore"]  # pylint: disable=no-member
            del Score.registry["NewRegressionScore"]  # pylint: disable=no-member

    def test_target_changing(self):
        data = Table("iris")
        w = self.widget  #: OWTestAndScore
        model = w.score_table.model

        w.n_folds = 2
        self.send_signal(self.widget.Inputs.train_data, data)
        self.send_signal(self.widget.Inputs.learner,
                         LogisticRegressionLearner(), 0, wait=5000)

        average_auc = float(model.item(0, 3).text())

        simulate.combobox_activate_item(w.controls.class_selection, "Iris-setosa")
        setosa_auc = float(model.item(0, 3).text())

        simulate.combobox_activate_item(w.controls.class_selection, "Iris-versicolor")
        versicolor_auc = float(model.item(0, 3).text())

        simulate.combobox_activate_item(w.controls.class_selection, "Iris-virginica")
        virginica_auc = float(model.item(0, 3).text())

        self.assertGreater(average_auc, versicolor_auc)
        self.assertGreater(average_auc, virginica_auc)
        self.assertLess(average_auc, setosa_auc)
        self.assertGreater(setosa_auc, versicolor_auc)
        self.assertGreater(setosa_auc, virginica_auc)

    def test_resort_on_data_change(self):
        iris = Table("iris")
        # one example is included from the other class
        # to keep F1 from complaining
        setosa = iris[:51]
        versicolor = iris[49:100]

        class SetosaLearner(Learner):
            def __call__(self, data):
                model = ConstantModel([1., 0, 0])
                model.domain = iris.domain
                return model

        class VersicolorLearner(Learner):
            def __call__(self, data):
                model = ConstantModel([0, 1., 0])
                model.domain = iris.domain
                return model

        # this is done manually to avoid multiple computations
        self.widget.resampling = 5
        self.widget.set_train_data(iris)
        self.widget.insert_learner(0, SetosaLearner())
        self.widget.insert_learner(1, VersicolorLearner())
        self.send_signal(self.widget.Inputs.test_data, setosa, wait=5000)

        self.widget.adjustSize()
        view = self.widget.score_table.view
        header = view.horizontalHeader()
        p = header.rect().center()
        # second visible header section (after 'Model')
        _, _, idx, *_ = (i for i in range(header.count())
                      if not header.isSectionHidden(i))
        p.setX(header.sectionPosition(idx) + 5)
        QTest.mouseClick(header.viewport(), Qt.LeftButton, pos=p)

        # Ensure that the click on header caused an ascending sort
        # Ascending sort means that wrong model should be listed first
        self.assertEqual(header.sortIndicatorOrder(), Qt.AscendingOrder)
        self.assertEqual(view.model().index(0, 0).data(), "versicolor")

        self.send_signal(self.widget.Inputs.test_data, versicolor, wait=5000)
        self.assertEqual(view.model().index(0, 0).data(), "setosa")

    def _retrieve_scores(self):
        w = self.widget
        model = w.score_table.model
        auc = model.item(0, 3).text()
        auc = float(auc) if auc != "" else None
        ca = float(model.item(0, 4).text())
        f1 = float(model.item(0, 5).text())
        precision = float(model.item(0, 6).text())
        recall = float(model.item(0, 7).text())
        return auc, ca, f1, precision, recall

    def _test_scores(self, train_data, test_data, learner, sampling, n_folds):
        w = self.widget  #: OWTestAndScore
        w.controls.resampling.buttons[sampling].click()
        if n_folds is not None:
            w.n_folds = n_folds

        self.send_signal(self.widget.Inputs.train_data, train_data)
        if test_data is not None:
            self.send_signal(self.widget.Inputs.test_data, test_data)
        self.send_signal(self.widget.Inputs.learner, learner, 0, wait=5000)
        return self._retrieve_scores()

    def test_scores_constant(self):
        table = Table.from_list(
            self.scores_domain,
            list(zip(*self.scores_table_values + [list("yyyn")]))
        )

        scores = self._test_scores(
            table, table[:3], ConstantLearner(),
            OWTestAndScore.TestOnTest, None
        )
        self.assertTupleEqual(scores[1:], (1, 1, 1, 1))

        # Sklearn, we love you.
        if parse_version(sklearn_version) < parse_version("1.6dev"):
            self.assertIsNone(scores[0])
        else:
            self.assertTrue(np.isnan(scores[0]))

    def test_scores_log_reg_overfitted(self):
        table = Table.from_list(
            self.scores_domain,
            list(zip(*self.scores_table_values + [list("yyyn")]))
        )

        self.assertTupleEqual(self._test_scores(
            table, table, LogisticRegressionLearner(),
            OWTestAndScore.TestOnTest, None),
            (1, 1, 1, 1, 1))

    def test_scores_log_reg_bad(self):
        table_train = Table.from_list(
            self.scores_domain,
            list(zip(*self.scores_table_values + [list("nnny")]))
        )
        table_test = Table.from_list(
            self.scores_domain,
            list(zip(*self.scores_table_values + [list("yyyn")]))
        )

        self.assertTupleEqual(self._test_scores(
            table_train, table_test, LogisticRegressionLearner(),
            OWTestAndScore.TestOnTest, None),
            (0, 0, 0, 0, 0))

    def test_scores_log_reg_bad2(self):
        table_train = Table.from_list(
            self.scores_domain,
            list(zip(*(self.scores_table_values + [list("nnyy")]))))
        table_test = Table.from_list(
            self.scores_domain,
            list(zip(*(self.scores_table_values + [list("yynn")]))))

        lr = LogisticRegressionLearner()
        np.testing.assert_almost_equal(
            self._test_scores(
                table_train, table_test, lr, OWTestAndScore.TestOnTest, None
            ),
            (0, 0.25, 0.2, 0.1666666, 0.25),
        )

    def test_scores_log_reg_advanced(self):
        table_train = Table.from_list(
            self.scores_domain,
            list(zip([1, 1, 1.23, 23.8, 5.], [1., 2., 3., 4., 3.], "yyynn"))
        )
        table_test = Table.from_list(
            self.scores_domain,
            list(zip([1, 1, 1.23, 23.8, 5.], [1., 2., 3., 4., 3.], "yynnn"))
        )

        lr = LogisticRegressionLearner()
        np.testing.assert_
        np.testing.assert_almost_equal(
            self._test_scores(
                table_train, table_test, lr, OWTestAndScore.TestOnTest, None
            ),
            (1, 0.8, 0.8, 13 / 15, 0.8)
        )

    def test_scores_cross_validation(self):
        """
        Test more than two classes and cross-validation
        """
        self.assertTrue(
            all(x >= y for x, y in zip(
                self._test_scores(
                    Table("iris")[::15], None, LogisticRegressionLearner(),
                    OWTestAndScore.KFold, 0),
                (0.8, 0.5, 0.5, 0.5, 0.5))))

    def test_no_stratification(self):
        w = self.widget
        w.cv_stratified = True
        self._test_scores(
            Table("zoo"), None, MajorityLearner(),
            OWTestAndScore.KFold, 2)
        self.assertTrue(w.Warning.cant_stratify.is_shown())

        w.controls.cv_stratified.click()
        self.assertFalse(w.Warning.cant_stratify.is_shown())

        w.controls.cv_stratified.click()
        self.assertTrue(w.Warning.cant_stratify.is_shown())

        w.controls.n_folds.setCurrentIndex(0)
        w.controls.n_folds.activated[int].emit(0)
        self.assertFalse(w.Warning.cant_stratify.is_shown())

        w.controls.n_folds.setCurrentIndex(2)
        w.controls.n_folds.activated[int].emit(2)
        self.assertTrue(w.Warning.cant_stratify.is_shown())

        self._test_scores(
            Table("iris"), None, MajorityLearner(), OWTestAndScore.KFold, 2)
        self.assertFalse(w.Warning.cant_stratify.is_shown())

        self._test_scores(
            Table("zoo"), None, MajorityLearner(), OWTestAndScore.KFold, 2)
        self.assertTrue(w.Warning.cant_stratify.is_shown())

        self._test_scores(
            Table("housing"), None, MeanLearner(), OWTestAndScore.KFold, 2)
        self.assertFalse(w.Warning.cant_stratify.is_shown())
        self.assertTrue(w.Information.cant_stratify_numeric.is_shown())

        w.controls.cv_stratified.click()
        self.assertFalse(w.Warning.cant_stratify.is_shown())

    def test_too_many_folds(self):
        w = self.widget
        w.controls.resampling.buttons[OWTestAndScore.KFold].click()
        w.n_folds = 3
        self.send_signal(w.Inputs.train_data, Table("zoo")[:8])
        self.send_signal(w.Inputs.learner, MajorityLearner(), 0, wait=5000)
        self.assertTrue(w.Error.too_many_folds.is_shown())

    def _set_comparison_score(self, score):
        w = self.widget
        control = w.controls.comparison_criterion
        control.setCurrentText(score)
        w.comparison_criterion = control.findText(score)

    def _set_three_majorities(self):
        w = self.widget
        data = Table("iris")[::15]
        self.send_signal(w.Inputs.train_data, data)
        for i, name in enumerate(["maja", "majb", "majc"]):
            learner = MajorityLearner()
            learner.name = name
            self.send_signal(w.Inputs.learner, learner, i)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)

    @patch("baycomp.two_on_single", Mock(wraps=baycomp.two_on_single))
    def test_comparison_requires_cv(self):
        w = self.widget
        self.send_signal(w.Inputs.train_data, Table("iris")[::15])

        w.comparison_criterion = 1
        rbs = w.controls.resampling.buttons

        self._set_three_majorities()
        baycomp.two_on_single.reset_mock()

        rbs[OWTestAndScore.KFold].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsNotNone(w.comparison_table.cellWidget(0, 1))
        self.assertTrue(w.compbox.isEnabled())
        baycomp.two_on_single.assert_called()
        baycomp.two_on_single.reset_mock()

        rbs[OWTestAndScore.LeaveOneOut].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsNone(w.comparison_table.cellWidget(0, 1))
        self.assertFalse(w.compbox.isEnabled())
        baycomp.two_on_single.assert_not_called()
        baycomp.two_on_single.reset_mock()

        rbs[OWTestAndScore.KFold].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsNotNone(w.comparison_table.cellWidget(0, 1))
        self.assertTrue(w.compbox.isEnabled())
        baycomp.two_on_single.assert_called()
        baycomp.two_on_single.reset_mock()

    def test_comparison_requires_multiple_models(self):
        w = self.widget
        rbs = w.controls.resampling.buttons

        self._set_three_majorities()
        w.comparison_criterion = 1

        rbs[OWTestAndScore.KFold].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertTrue(w.comparison_table.isEnabled())

        self.send_signal(w.Inputs.learner, None, 1)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertTrue(w.comparison_table.isEnabled())

        self.send_signal(w.Inputs.learner, None, 2)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertFalse(w.comparison_table.isEnabled())

        rbs[OWTestAndScore.LeaveOneOut].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertFalse(w.comparison_table.isEnabled())

        learner = MajorityLearner()
        learner.name = "majd"
        self.send_signal(w.Inputs.learner, learner, 1)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertFalse(w.comparison_table.isEnabled())

        rbs[OWTestAndScore.KFold].click()
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertTrue(w.comparison_table.isEnabled())

    def test_comparison_bad_slots(self):
        w = self.widget
        self._set_three_majorities()
        self._set_comparison_score("Classification accuracy")
        self.send_signal(w.Inputs.learner, BadLearner(), 2, wait=5000)
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)
        self.assertIsNotNone(w.comparison_table.cellWidget(0, 1))
        self.assertIsNone(w.comparison_table.cellWidget(0, 2))
        self.assertEqual(len(w._successful_slots()), 2)

    def test_comparison_bad_scores(self):
        w = self.widget
        self._set_three_majorities()
        self._set_comparison_score("Classification accuracy")
        self.get_output(self.widget.Outputs.evaluations_results, wait=5000)

        score_calls = -1

        def fail_on_first(*_, **_2):
            nonlocal score_calls
            score_calls += 1
            return 1 / score_calls

        with patch.object(scoring.CA, "compute_score", new=fail_on_first):
            w.update_comparison_table()

            self.assertIsNone(w.comparison_table.cellWidget(0, 1))
            self.assertIsNone(w.comparison_table.cellWidget(0, 2))
            self.assertIsNone(w.comparison_table.cellWidget(1, 0))
            self.assertIsNone(w.comparison_table.cellWidget(2, 0))
            self.assertIsNotNone(w.comparison_table.cellWidget(1, 2))
            self.assertIsNotNone(w.comparison_table.cellWidget(2, 1))
            self.assertTrue(w.Warning.scores_not_computed.is_shown())

        score_calls = -1
        with patch.object(scoring.CA, "compute_score", new=fail_on_first):
            slots = w._successful_slots()
            self.assertEqual(len(slots), 3)
            scores = w._scores_by_folds(slots)
            self.assertIsNone(scores[0])
            self.assertEqual(scores[1][0], 1)
            self.assertAlmostEqual(scores[2][0], 1 / 6)

    def test_comparison_binary_score(self):
        # false warning at call_arg.kwargs
        # pylint: disable=unpacking-non-sequence
        w = self.widget
        self._set_three_majorities()
        self._set_comparison_score("F1")
        f1mock = Mock(wraps=scoring.F1.compute_score)

        iris = Table("iris")
        with patch.object(scoring.F1, "compute_score", f1mock):
            simulate.combobox_activate_item(w.controls.class_selection,
                                            iris.domain.class_var.values[1])
            _, kwargs = f1mock.call_args
            self.assertEqual(kwargs["target"], 1)
            self.assertFalse("average" in kwargs)

            simulate.combobox_activate_item(w.controls.class_selection,
                                            iris.domain.class_var.values[2])
            _, kwargs = f1mock.call_args
            self.assertEqual(kwargs["target"], 2)
            self.assertFalse("average" in kwargs)

            simulate.combobox_activate_item(w.controls.class_selection,
                                            OWTestAndScore.TARGET_AVERAGE)
            _, kwargs = f1mock.call_args
            self.assertEqual(kwargs["average"], "weighted")
            self.assertFalse("target" in kwargs)

    def test_fill_table(self):
        w = self.widget
        self._set_three_majorities()
        scores = [object(), object(), object()]
        slots = w._successful_slots()

        def probs(p1, p2, rope):
            p1 += 1
            p2 += 1
            norm = p1 + p2 + rope * (p1 + p2)
            if rope == 0:
                return p1 / norm, p2 / norm
            else:
                return p1 / norm, rope / norm, p2 / norm

        def two_on_single(res1, res2, rope=0):
            return probs(scores.index(res1), scores.index(res2), rope)

        with patch("baycomp.two_on_single", new=two_on_single):
            for w.use_rope, w.rope in ((True, 0), (False, 0.1)):
                w._fill_table(slots, scores)
                for row in range(3):
                    for col in range(3):
                        if row == col:
                            continue
                        label = w.comparison_table.cellWidget(row, col)
                        self.assertEqual(label.text(),
                                         f"{(row + 1) / (row + col + 2):.3f}")
                        self.assertIn(f"{(row + 1) / (row + col + 2):.3f}",
                                      label.toolTip())

            w.use_rope = True
            w.rope = 0.25
            w._fill_table(slots, scores)
            for row in range(3):
                for col in range(3):
                    if row == col:
                        continue
                    label = w.comparison_table.cellWidget(row, col)
                    for text in (label.text(), label.toolTip()):
                        self.assertIn(f"{probs(row, col, w.rope)[0]:.3f}", text)
                        self.assertIn(f"{probs(row, col, w.rope)[1]:.3f}", text)

    def test_nan_on_comparison(self):
        w = self.widget
        w.use_rope = True
        self._set_three_majorities()
        scores = [object(), object(), object()]
        slots = w._successful_slots()

        def two_on_single(_1, _2, rope=0):
            if rope:
                return np.nan, np.nan, np.nan
            else:
                return np.nan, np.nan

        with patch("baycomp.two_on_single", new=two_on_single):
            for w.rope in (0, 0.1):
                w._fill_table(slots, scores)
                label = w.comparison_table.cellWidget(1, 0)
                self.assertEqual(label.text(), "NA")

    def test_unique_output_domain(self):
        data = possible_duplicate_table('random forest')
        self.send_signal(self.widget.Inputs.train_data, data)
        self.send_signal(self.widget.Inputs.learner, RandomForestLearner(), 0)
        output = self.get_output(self.widget.Outputs.predictions)
        self.assertEqual(output.domain.metas[0].name, 'random forest (1)')

    def test_copy_to_clipboard(self):
        self.send_signal(self.widget.Inputs.train_data, Table("iris"))
        self.send_signal(self.widget.Inputs.learner, RandomForestLearner(), 0)
        self.wait_until_finished()
        view = self.widget.score_table.view
        model = self.widget.score_table.model
        selection_model = view.selectionModel()
        selection_model.select(model.index(0, 0),
                               selection_model.Select | selection_model.Rows)
        self.widget.copy_to_clipboard()
        clipboard_text = QApplication.clipboard().text()
        # Tests appear to register additional scorers, so we clip the list
        # to what we know to be there and visible
        clipboard_text = "\t".join(clipboard_text.split("\t")[:6]).strip()
        view_text = "\t".join([str(model.data(model.index(0, i)))
                               for i in (0, 3, 4, 5, 6, 7)]).strip()
        self.assertEqual(clipboard_text, view_text)

    def test_multi_target_input(self):
        class NewScorer(Score):
            class_types = (
                ContinuousVariable,
                DiscreteVariable,
            )

            @staticmethod
            def is_compatible(domain: Domain) -> bool:
                return True

            def compute_score(self, results):
                return [0.75]

        domain = Domain([ContinuousVariable('var1')],
                        class_vars=[
                            ContinuousVariable('c1'),
                            DiscreteVariable('c2', values=('no', 'yes'))
                        ])
        data = Table.from_list(domain, [[1, 5, 0], [2, 10, 1], [2, 10, 1]])

        mock_model = Mock(spec=Model, return_value=np.asarray([0.2, 0.1, 0.2]))
        mock_model.name = 'Mockery'
        mock_model.domain = domain
        mock_learner = Mock(spec=Learner, return_value=mock_model)
        mock_learner.name = 'Mockery'

        widget = self.create_widget(OWTestAndScore)
        widget.resampling = OWTestAndScore.TestOnTrain
        self.send_signal(widget.Inputs.train_data, data)
        self.send_signal(widget.Inputs.learner, MajorityLearner(), 0)
        self.send_signal(widget.Inputs.learner, mock_learner, 1)
        _ = self.get_output(widget.Outputs.evaluations_results, wait=5000)
        self.assertTrue(len(widget.scorers) == 1)
        self.assertTrue(NewScorer in widget.scorers)
        self.assertTrue(len(widget._successful_slots()) == 1)

    def test_multiple_targets_pls(self):
        housing = Table("housing")
        class_vars = [housing.domain.class_var, housing.domain.attributes[0]]
        domain = Domain(housing.domain.attributes[1:], class_vars=class_vars)
        multiple_targets_data = housing.transform(domain)

        self.widget.error = Mock()
        self.send_signal(self.widget.Inputs.train_data, multiple_targets_data)
        self.send_signal(self.widget.Inputs.learner, PLSRegressionLearner())
        self.wait_until_finished()
        self.assertIn("Multiple targets are not supported.",
                      self.widget.error.call_args[0][0])


class TestHelpers(unittest.TestCase):
    def test_results_one_vs_rest(self):
        data = Table(test_filename("datasets/lenses.tab"))
        learners = [MajorityLearner()]
        res = TestOnTestData()(data[1::2], data[::2], learners=learners)
        r1 = results_one_vs_rest(res, pos_index=0)
        r2 = results_one_vs_rest(res, pos_index=1)
        r3 = results_one_vs_rest(res, pos_index=2)

        np.testing.assert_almost_equal(np.sum(r1.probabilities, axis=2), 1.0)
        np.testing.assert_almost_equal(np.sum(r2.probabilities, axis=2), 1.0)
        np.testing.assert_almost_equal(np.sum(r3.probabilities, axis=2), 1.0)

        np.testing.assert_almost_equal(
            r1.probabilities[:, :, 1] +
            r2.probabilities[:, :, 1] +
            r3.probabilities[:, :, 1],
            1.0
        )
        self.assertEqual(r1.folds, res.folds)
        self.assertEqual(r2.folds, res.folds)
        self.assertEqual(r3.folds, res.folds)

        np.testing.assert_equal(r1.row_indices, res.row_indices)
        np.testing.assert_equal(r2.row_indices, res.row_indices)
        np.testing.assert_equal(r3.row_indices, res.row_indices)


if __name__ == "__main__":
    unittest.main()



from typing import Optional, Callable, Collection, Sequence

import numpy as np
from AnyQt.QtCore import QPointF, Qt, QSize
from AnyQt.QtGui import QStandardItemModel, QStandardItem, \
    QPainter, QFontMetrics
from AnyQt.QtWidgets import QGraphicsSceneHelpEvent, QToolTip, \
    QGridLayout, QSizePolicy, QWidget

import pyqtgraph as pg

from orangewidget.utils.itemmodels import signal_blocking
from orangewidget.utils.visual_settings_dlg import VisualSettingsDialog, \
    KeyType, ValueType

from Orange.base import Learner
from Orange.data import Table
from Orange.evaluation import CrossValidation, TestOnTrainingData, Results
from Orange.evaluation.scoring import Score, AUC, R2
from Orange.modelling import Fitter
from Orange.util import dummy_callback
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils import userinput
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin, TaskState
from Orange.widgets.utils.multi_target import check_multiple_targets_input
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.owscatterplotgraph import LegendItem
from Orange.widgets.visualize.utils.customizableplot import \
    CommonParameterSetter, Updater
from Orange.widgets.visualize.utils.plotutils import PlotWidget, \
    HelpEventDelegate
from Orange.widgets.widget import OWWidget, Input, Msg

N_FOLD = 7
MIN_MAX_SPIN = 100000
ScoreType = tuple[int, tuple[float, float]]
# scores, score name, label
FitterResults = tuple[list[ScoreType], str, str]


def _validate(
        data: Table,
        learner: Learner,
        scorer: type[Score]
) -> tuple[float, float]:
    res: Results = TestOnTrainingData()(data, [learner],
                                        suppresses_exceptions=False)
    res_cv: Results = CrossValidation(k=N_FOLD)(data, [learner],
                                                suppresses_exceptions=False)
    # pylint: disable=unsubscriptable-object
    return scorer(res)[0], scorer(res_cv)[0]


def _search(
        data: Table,
        learner: Learner,
        fitted_parameter_props: Learner.FittedParameter,
        initial_parameters: dict[str, int],
        steps: Collection[int],
        progress_callback: Callable = dummy_callback
) -> FitterResults:
    progress_callback(0, "Calculating...")
    scores = []
    scorer = AUC if data.domain.has_discrete_class else R2
    name = fitted_parameter_props.name
    for i, value in enumerate(steps):
        progress_callback(i / len(steps))
        params = initial_parameters.copy()
        params[name] = value
        result = _validate(data, type(learner)(**params), scorer)
        scores.append((value, result))
    return scores, scorer.name, fitted_parameter_props.label


def run(
        data: Table,
        learner: Learner,
        fitted_parameter_props: Learner.FittedParameter,
        initial_parameters: dict[str, int],
        steps: Collection[int],
        state: TaskState
) -> FitterResults:
    def callback(i: float, status: str = ""):
        state.set_progress_value(i * 100)
        if status:
            state.set_status(status)
        if state.is_interruption_requested():
            # pylint: disable=broad-exception-raised
            raise Exception

    return _search(data, learner, fitted_parameter_props, initial_parameters,
                   steps, callback)


class ParameterSetter(CommonParameterSetter):
    GRID_LABEL, SHOW_GRID_LABEL = "Gridlines", "Show"
    DEFAULT_ALPHA_GRID, DEFAULT_SHOW_GRID = 80, True

    def __init__(self, master):
        self.grid_settings: Optional[dict] = None
        self.master: FitterPlot = master
        super().__init__()

    def update_setters(self):
        self.grid_settings = {
            Updater.ALPHA_LABEL: self.DEFAULT_ALPHA_GRID,
            self.SHOW_GRID_LABEL: self.DEFAULT_SHOW_GRID,
        }

        self.initial_settings = {
            self.LABELS_BOX: {
                self.FONT_FAMILY_LABEL: self.FONT_FAMILY_SETTING,
                self.AXIS_TITLE_LABEL: self.FONT_SETTING,
                self.AXIS_TICKS_LABEL: self.FONT_SETTING,
                self.LEGEND_LABEL: self.FONT_SETTING,
            },
            self.PLOT_BOX: {
                self.GRID_LABEL: {
                    self.SHOW_GRID_LABEL: (None, True),
                    Updater.ALPHA_LABEL: (range(0, 255, 5),
                                          self.DEFAULT_ALPHA_GRID),
                },
            },
        }

        def update_grid(**settings):
            self.grid_settings.update(**settings)
            self.master.showGrid(
                x=False, y=self.grid_settings[self.SHOW_GRID_LABEL],
                alpha=self.grid_settings[Updater.ALPHA_LABEL] / 255)

        self._setters[self.PLOT_BOX] = {self.GRID_LABEL: update_grid}

    @property
    def axis_items(self):
        return [value["item"] for value in
                self.master.getPlotItem().axes.values()]

    @property
    def legend_items(self):
        return self.master.legend.items


class FitterPlot(PlotWidget):
    BAR_WIDTH = 0.4

    def __init__(self):
        super().__init__(enableMenu=False)
        self.__bar_item_tr: Optional[pg.BarGraphItem] = None
        self.__bar_item_cv: Optional[pg.BarGraphItem] = None
        self.__data: Optional[list[ScoreType]] = None
        self.legend = self._create_legend()
        self.parameter_setter = ParameterSetter(self)
        self.setMouseEnabled(False, False)
        self.hideButtons()

        self.showGrid(x=False, y=self.parameter_setter.DEFAULT_SHOW_GRID,
                      alpha=self.parameter_setter.DEFAULT_ALPHA_GRID / 255)

        self.tooltip_delegate = HelpEventDelegate(self.help_event)
        self.scene().installEventFilter(self.tooltip_delegate)

    def _create_legend(self) -> LegendItem:
        legend = LegendItem()
        legend.setParentItem(self.getViewBox())
        legend.anchor((1, 1), (1, 1), offset=(-5, -5))
        legend.hide()
        return legend

    def clear_all(self):
        self.clear()
        self.__bar_item_tr = None
        self.__bar_item_cv = None
        self.__data = None
        self.setLabel(axis="bottom", text=None)
        self.setLabel(axis="left", text=None)
        self.getAxis("bottom").setTicks(None)

    def set_data(
            self,
            scores: list[ScoreType],
            score_name: str,
            parameter_name: str
    ):
        self.__data = scores
        self.clear()
        self.setLabel(axis="bottom", text=parameter_name)
        self.setLabel(axis="left", text=score_name)

        ticks = [[(i, str(val)) for i, (val, _)
                  in enumerate(scores)]]
        self.getAxis("bottom").setTicks(ticks)

        brush_tr = "#6fa255"
        brush_cv = "#3a78b6"
        pen = pg.mkPen("#333")
        kwargs = {"pen": pen, "width": self.BAR_WIDTH}
        bar_item_tr = pg.BarGraphItem(x=np.arange(len(scores)) - 0.2,
                                      height=[(s[0]) for _, s in scores],
                                      brush=brush_tr, **kwargs)
        bar_item_cv = pg.BarGraphItem(x=np.arange(len(scores)) + 0.2,
                                      height=[(s[1]) for _, s in scores],
                                      brush=brush_cv, **kwargs)
        self.addItem(bar_item_tr)
        self.addItem(bar_item_cv)
        self.__bar_item_tr = bar_item_tr
        self.__bar_item_cv = bar_item_cv

        self.legend.clear()
        kwargs = {"pen": pen, "symbol": "s"}
        scatter_item_tr = pg.ScatterPlotItem(brush=brush_tr, **kwargs)
        scatter_item_cv = pg.ScatterPlotItem(brush=brush_cv, **kwargs)
        self.legend.addItem(scatter_item_tr, "Train")
        self.legend.addItem(scatter_item_cv, "CV")
        Updater.update_legend_font(self.legend.items,
                                   **self.parameter_setter.legend_settings)
        self.legend.show()

    def help_event(self, ev: QGraphicsSceneHelpEvent) -> bool:
        if self.__bar_item_tr is None:
            return False

        pos = self.__bar_item_tr.mapFromScene(ev.scenePos())
        index = self.__get_index_at(pos)
        text = ""
        if index is not None:
            _, scores = self.__data[index]
            text = "<table align=left>" \
                   "<tr>" \
                   "<td><b>Train:</b></td>" \
                   f"<td>{round(scores[0], 3)}</td>" \
                   "</tr><tr>" \
                   "<td><b>CV:</b></td>" \
                   f"<td>{round(scores[1], 3)}</td>" \
                   "</tr>" \
                   "</table>"
        if text:
            QToolTip.showText(ev.screenPos(), text, widget=self)
            return True
        else:
            return False

    def __get_index_at(self, point: QPointF) -> Optional[int]:
        x = point.x()
        index = round(x)
        # pylint: disable=unsubscriptable-object
        heights_tr: list = self.__bar_item_tr.opts["height"]
        heights_cv: list = self.__bar_item_cv.opts["height"]
        if 0 <= index < len(heights_tr) and abs(index - x) <= self.BAR_WIDTH:
            if index > x and 0 <= point.y() <= heights_tr[index]:
                return index
            if x > index and 0 <= point.y() <= heights_cv[index]:
                return index
        return None


class RangePreview(QWidget):
    def __init__(self):
        super().__init__()
        font = self.font()
        font.setPointSize(font.pointSize() - 3)
        self.setFont(font)

        self.__steps: Optional[Sequence[int]] = None
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)

    def minimumSizeHint(self):
        return QSize(1, 20)

    def set_steps(self, steps: Optional[Sequence[int]]):
        self.__steps = steps
        self.update()

    def steps(self):
        return self.__steps

    def paintEvent(self, _):
        if not self.__steps:
            return
        painter = QPainter(self)
        metrics = QFontMetrics(self.font())
        style = self.style()
        rect = self.rect()

        # Indent by the width of the radio button indicator
        rect.adjust(style.pixelMetric(style.PM_IndicatorWidth)
                    + style.pixelMetric(style.PM_CheckBoxLabelSpacing), 0, 0, 0)

        last_text = f"{self.__steps[-1]}"
        if len(self.__steps) > 1:
            last_text = ", " + last_text
        last_width = metrics.horizontalAdvance(last_text)

        elided_text = metrics.elidedText(
            "Steps: " + ", ".join(map(str, self.__steps[:-1])),
            Qt.ElideRight, rect.width() - last_width)
        elided_width = metrics.horizontalAdvance(elided_text)

        # Right-align by indenting by the underflow width
        rect.adjust(rect.width() - elided_width - last_width, 0, 0, 0)

        painter.drawText(rect, Qt.AlignLeft, elided_text)
        rect.adjust(elided_width, 0, 0, 0)
        painter.drawText(rect, Qt.AlignLeft, last_text)


class OWParameterFitter(OWWidget, ConcurrentWidgetMixin):
    name = "Parameter Fitter"
    description = "Fit learner for various values of fitting parameter."
    icon = "icons/ParameterFitter.svg"
    priority = 1110

    visual_settings = Setting({}, schema_only=True)
    graph_name = "graph.plotItem"

    class Inputs:
        data = Input("Data", Table)
        learner = Input("Learner", Learner)

    DEFAULT_PARAMETER_INDEX = 0
    DEFAULT_MINIMUM = 1
    DEFAULT_MAXIMUM = 9
    parameter_index = Setting(DEFAULT_PARAMETER_INDEX, schema_only=True)
    FROM_RANGE, MANUAL = range(2)
    type: int = Setting(FROM_RANGE)
    minimum: int = Setting(DEFAULT_MINIMUM, schema_only=True)
    maximum: int = Setting(DEFAULT_MAXIMUM, schema_only=True)
    manual_steps: str = Setting("", schema_only=True)
    auto_commit = Setting(True)

    class Error(OWWidget.Error):
        unknown_err = Msg("{}")
        not_enough_data = Msg(f"At least {N_FOLD} instances are needed.")
        incompatible_learner = Msg("{}")
        manual_steps_error = Msg("Invalid values for '{}': {}")
        min_max_error = Msg("Minimum must be less than maximum.")
        missing_target = Msg("Data has no target.")

    class Warning(OWWidget.Warning):
        no_parameters = Msg("{} has no parameters to fit.")

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)
        self._data: Optional[Table] = None
        self._learner: Optional[Learner] = None
        self.__parameters_model = QStandardItemModel()
        self.__initialize_settings = False

        self.setup_gui()
        VisualSettingsDialog(
            self, self.graph.parameter_setter.initial_settings
        )

    def setup_gui(self):
        self._add_plot()
        self._add_controls()

    def _add_plot(self):
        # This is a part of __init__
        # pylint: disable=attribute-defined-outside-init
        box = gui.vBox(self.mainArea)
        self.graph = FitterPlot()
        box.layout().addWidget(self.graph)

    def _add_controls(self):
        # This is a part of __init__
        # pylint: disable=attribute-defined-outside-init
        layout = QGridLayout()
        gui.widgetBox(self.controlArea, "Settings", orientation=layout)
        self.__combo = gui.comboBox(None, self, "parameter_index",
                                    model=self.__parameters_model,
                                    callback=self.__on_parameter_changed)
        layout.addWidget(self.__combo, 0, 0, 1, 2)

        buttons = gui.radioButtons(None, self, "type",
                                   callback=self.__on_type_changed)
        button = gui.appendRadioButton(buttons, "Range:")
        layout.addWidget(button, 1, 0)

        # pylint: disable=use-dict-literal
        kw = dict(minv=-MIN_MAX_SPIN, maxv=MIN_MAX_SPIN,
                  alignment=Qt.AlignRight,
                  callback=self.__on_min_max_changed)
        box = gui.hBox(None)
        self.__spin_min = gui.spin(box, self, "minimum", label="From:", **kw)
        layout.addWidget(box, 1, 1)

        box = gui.hBox(None)
        self.__spin_max = gui.spin(box, self, "maximum", label="To:", **kw)
        layout.addWidget(box, 2, 1)

        self.range_preview = RangePreview()
        layout.addWidget(self.range_preview, 3, 0, 1, 2)

        gui.appendRadioButton(buttons, "Manual:")
        layout.addWidget(buttons, 4, 0)
        self.edit = gui.lineEdit(None, self, "manual_steps",
                                 placeholderText="e.g. 10, 20, ..., 50",
                                 callback=self.__on_manual_changed)
        layout.addWidget(self.edit, 4, 1)

        # gui.lineEdit's connect does not call the callback on return pressed
        # if the line hasn't changed.
        @self.edit.returnPressed.connect
        def _():
            if self.type != self.MANUAL:
                self.type = self.MANUAL
                self.__on_type_changed()

        gui.rubber(self.controlArea)

        gui.auto_apply(self.buttonsArea, self, "auto_commit")

        self._update_preview()

    def __on_type_changed(self):
        self._settings_changed()

    def __on_parameter_changed(self):
        self.__initialize_settings = True
        self._set_range_controls(self.fitted_parameters[self.parameter_index])
        self._settings_changed()

    def __on_min_max_changed(self):
        self.type = self.FROM_RANGE
        self._settings_changed()

    def __on_manual_changed(self):
        self.type = self.MANUAL
        self._settings_changed()

    def _settings_changed(self):
        self._update_preview()
        self.commit.deferred()

    @property
    def fitted_parameters(self) -> list:
        if not self._learner:
            return []
        return self._learner.fitted_parameters

    @property
    def initial_parameters(self) -> dict:
        if not self._learner:
            return {}
        if isinstance(self._learner, Fitter):
            return self._learner.get_params(self._data or "classification")
        return self._learner.params

    @property
    def steps(self) -> tuple[int, ...]:
        self.Error.min_max_error.clear()
        self.Error.manual_steps_error.clear()

        if self.type == self.FROM_RANGE:
            return self._steps_from_range()
        else:
            return self._steps_from_manual()

    def _steps_from_range(self) -> tuple[int, ...]:
        if self.maximum < self.minimum:
            self.Error.min_max_error()
            return ()

        if self.minimum == self.maximum:
            return (self.minimum, )

        diff = self.maximum - self.minimum
        # This should give between 10 and 15 steps
        exp = max(0, int(np.ceil(np.log10(diff / 1.5))) - 1)
        step = int(10 ** exp)
        return (self.minimum,
                *range((self.minimum // step + 1) * step, self.maximum, step),
                self.maximum)

    def _steps_from_manual(self) -> tuple[int, ...]:
        param = self.fitted_parameters[self.parameter_index]
        try:
            steps = userinput.numbers_from_list(
                self.manual_steps, int, param.min, param.max)
        except ValueError as ex:
            self.Error.manual_steps_error(param.label, ex)
            return ()
        if steps and "..." not in self.manual_steps:
            self.manual_steps = ", ".join(map(str, steps))
        return steps

    @Inputs.data
    @check_multiple_targets_input
    def set_data(self, data: Optional[Table]):
        self.Error.not_enough_data.clear()
        self.Error.missing_target.clear()
        self._data = data
        if self._data and len(self._data) < N_FOLD:
            self.Error.not_enough_data()
            self._data = None
        if self._data and len(self._data.domain.class_vars) < 1:
            self.Error.missing_target()
            self._data = None

    @Inputs.learner
    def set_learner(self, learner: Optional[Learner]):
        self.Warning.clear()
        self.Error.manual_steps_error.clear()
        self.Error.min_max_error.clear()
        self.__parameters_model.clear()

        if not learner:
            self.__initialize_settings = False
            # reset spin controls
            ars = (None, None, int, None, None)
            self._set_range_controls(Learner.FittedParameter(*ars))

        elif self._learner:
            self.__initialize_settings = \
                learner.fitted_parameters != self.fitted_parameters

        else:
            # changed by user or opened workflow
            self.__initialize_settings = \
                self.parameter_index == self.DEFAULT_PARAMETER_INDEX and \
                self.minimum == self.DEFAULT_MINIMUM and \
                self.maximum == self.DEFAULT_MAXIMUM

        self._learner = learner
        if self._learner is None:
            return

        for param in self.fitted_parameters:
            item = QStandardItem(param.label)
            self.__parameters_model.appendRow(item)

        if not self.fitted_parameters:
            self.Warning.no_parameters(self._learner.name)
        else:
            if self.__initialize_settings:
                self.parameter_index = 0
            else:
                self.__combo.setCurrentIndex(self.parameter_index)
            self._set_range_controls(
                self.fitted_parameters[self.parameter_index])

        self._update_preview()

    def handleNewSignals(self):
        self.Error.unknown_err.clear()
        self.Error.incompatible_learner.clear()
        self.clear()

        if not self._data or not self._learner:
            return

        reason = self._learner.incompatibility_reason(self._data.domain)
        if reason:
            self.Error.incompatible_learner(reason)
            return

        self.commit.now()

    def _set_range_controls(self, param: Learner.FittedParameter):
        assert param.type == int, \
            "The widget currently supports only int parameters"

        # Block signals to avoid changing `self.type`
        with signal_blocking(self.__spin_min), signal_blocking(self.__spin_max):
            if param.min is not None:
                self.__spin_min.setMinimum(param.min)
                self.__spin_max.setMinimum(param.min)
                self.minimum = param.min if self.__initialize_settings else \
                    max(self.minimum, param.min)
            else:
                self.__spin_min.setMinimum(-MIN_MAX_SPIN)
                self.__spin_max.setMinimum(-MIN_MAX_SPIN)
                if self.__initialize_settings:
                    self.minimum = self.initial_parameters[param.name]
            if param.max is not None:
                self.__spin_min.setMaximum(param.max)
                self.__spin_max.setMaximum(param.max)
                if self.__initialize_settings:
                    self.maximum = param.max
                self.maximum = param.max if self.__initialize_settings else \
                    min(self.maximum, param.max)
            else:
                self.__spin_min.setMaximum(MIN_MAX_SPIN)
                self.__spin_max.setMaximum(MIN_MAX_SPIN)
                if self.__initialize_settings:
                    self.maximum = self.initial_parameters[param.name]
            self.__initialize_settings = False

        tip = "Enter a list of values"
        if param.min is not None:
            if param.max is not None:
                self.edit.setToolTip(f"{tip} between {param.min} and {param.max}.")
            else:
                self.edit.setToolTip(f"{tip} greater or equal to {param.min}.")
        elif param.max is not None:
            self.edit.setToolTip(f"{tip} smaller or equal to {param.max}.")
        else:
            self.edit.setToolTip("")

    def _update_preview(self):
        if self.type == self.FROM_RANGE:
            self.range_preview.set_steps(self.steps)
        else:
            self.range_preview.set_steps(None)

    def clear(self):
        self.cancel()
        self.graph.clear_all()

    @gui.deferred
    def commit(self):
        self.graph.clear_all()
        if self._data is None or self._learner is None or \
                not self.fitted_parameters or not self.steps:
            return
        self.start(run, self._data, self._learner,
                   self.fitted_parameters[self.parameter_index],
                   self.initial_parameters, self.steps)

    def on_done(self, result: FitterResults):
        self.graph.set_data(*result)

    def on_exception(self, ex: Exception):
        self.Error.unknown_err(ex)

    def on_partial_result(self, _):
        pass

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def send_report(self):
        if self._data is None or self._learner is None \
                or not self.fitted_parameters:
            return
        parameter = self.fitted_parameters[self.parameter_index].label
        self.report_items("Settings",
                          [("Parameter", parameter),
                           ("Range", ", ".join(map(str, self.steps)))])
        self.report_name("Plot")
        self.report_plot()

    def set_visual_settings(self, key: KeyType, value: ValueType):
        self.graph.parameter_setter.set_parameter(key, value)
        # pylint: disable=unsupported-assignment-operation
        self.visual_settings[key] = value


if __name__ == "__main__":
    from Orange.regression import PLSRegressionLearner

    WidgetPreview(OWParameterFitter).run(
        set_data=Table("housing"), set_learner=PLSRegressionLearner())



"""
========
Evaluate
========

Evaluating models.

"""

NAME = "Evaluate"

ID = "orange.widgets.evaluate"

DESCRIPTION = "Evaluate model performance"

BACKGROUND = "#C3F3F3"

ICON = "icons/Category-Evaluate.svg"

PRIORITY = 5
